/*
 * 	libglacier.c - Backend C library for Glacier
 *
 * 	This file is part of Glacier.
 *
 * 	Glacier is free software: you can redistribute it and/or modify it under the terms of the
 * 	GNU General Public License as published by the Free Software Foundation, either
 * 	version 3 of the License, or (at your option) any later version.
 *
 * 	Glacier is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * 	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * 	PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License along with Glacier. If
 * 	not, see <https://www.gnu.org/licenses/>.
 */

#define _POSIX_C_SOURCE 200809L

#include <color.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <libconfig.h>
#include <locale.h>
#include <openssl/evp.h>
#include <signal.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>  /* For strcasecmp */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <wchar.h>

#include "config.h"

/* Buffer and size constants */
#define BUFFER_SIZE 1024
#define MAX_CHILDREN 64
#define MAX_SIZE 256
#define PATH_MAX_SIZE 512
#define MAX_RECURSION_DEPTH 100
#define DEFAULT_PERMISSIONS 0750

/* Define LG_VERBOSE if not defined elsewhere */
#ifndef LG_VERBOSE
#define LG_VERBOSE 0
#endif

typedef unsigned int uint;
typedef unsigned char uchar;

/*
 * Global Variables
 * 
 * Descriptions are given as comments after the variable declaration.
 * DEFINED IN: globals.h
 *
 */

config_t cfg; /* Context for libconfig */
config_setting_t *setting = NULL; /* Pointer for setting */

/* Con:figuration variables with default values */
int GLACIER_ALLOW_SERVICES = 0; /* Declaration of GLACIER_ALLOW_SERVICES as given in glacier.cfg */
char *GLACIER_ALLOWED_LICENSES = NULL; /* Declaration of GLACIER_ALLOWED_LICENSES as given in glacier.cfg */
int GLACIER_DO_INT_CHECK = 1; /* Declaration of GLACIER_DO_INT_CHECK as given in glacier.cfg */
int GLACIER_VERBOSE = 0; /* Declaration of GLACIER_VERBOSE as given in glacier.cfg */

/* Profile variables */
const char *GLACIER_REPO = NULL; /* Declaration of GLACIER_REPO as defined in profile.cfg */
const char *GLACIER_ARCH = NULL; /* Declaration of GLACIER_ARCH as defined in profile.cfg */
const char *GLACIER_TARGET = NULL; /* Declaration of GLACIER_TARGET as defined in profile.cfg */
const char *GLACIER_LOCALDB = NULL; /* Declaration of GLACIER_LOCALDB as defined in profile.cfg */
const char *GLACIER_SYSTEM_PROFILE = NULL; /* Declaration of GLACIER_SYSTEM_PROFILE as defined in profile.cfg */

/* Required runtime files */
const char *runtime_files[] = { 
	"/etc/glacier.cfg",
	"/etc/glacier/call-hooks",
	"/etc/make.conf"
	};

FILE *expected_hash; /* File pointer for expected hash */
FILE *pkg; /* File pointer for package hash */

/*
 * END GLOBAL VARIABLES
 */

/*
 * node
 *
 * DESCRIPTION: Definition of node type.
 * DEFINED IN: data.h
 *
 */

struct node {
	char *data;
	struct node *children[MAX_CHILDREN];
	int numChildren;
	struct node *left;
	struct node *right;
};

/*
 * gl_infolog
 *
 * DESCRIPTION: Output a stylized info message.
 * PARAMETERS: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
gl_infolog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	printf(COL_BLUE "[i]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * gl_warnlog
 *
 * DESCRIPTION: Output a stylized warning message.
 * Parameters: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
gl_warnlog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	printf(COL_YELLOW "[!]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * gl_errlog
 * 
 * DESCRIPTION: Output a stylized error message.
 * PARAMETERS: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
gl_errlog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	fprintf(stderr, COL_RED "[x]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * gl_successlog
 * 
 * DESCRIPTION: Output a stylized success message.
 * PARAMETERS: char MSG[]
 * DEFINED IN: log.h
 *
 */

void
gl_successlog(char MSG[])
{
	if (MSG == NULL) {
		return;
	}

	printf(COL_GREEN "[*]" COL_RESET " %s\n", MSG);
	return;
}

/*
 * gl_runtime_exists
 *
 * DESCRIPTION: Check if necesary runtime files exist.
 * PARAMETERS: None.
 * DEFINED IN: runtime.h
 *
 */

int
gl_runtime_exists(void)
{
	int f;

	for (f = 0; f < sizeof(runtime_files); f++) {
		if (access(runtime_files[f], F_OK) == 0) {
			continue;
		} else {
			/* might keep this message, might not, idk */
			/* errlog("Runtime files are missing, cannot continue with operation.");
			printf(COL_RED "[x]" COL_RESET " The following files are missing:\n");
			printf(COL_RED "[x]" COL_RESET " \t%s\n", runtime_files[f]); */
			return 0;
		}
	}
	return 1;
}

/*
 * gl_is_process_root
 *
 * DESCRIPTION: Check if process is running as root.
 * PARAMETERS: char MSG[]
 * DEFINED IN: runtime.h
 *
 */

int
gl_is_process_root(void)
{
	if (getuid() != 0) {
		return 0; /* process is not running as root */
	}
	else {
		return 1; /* process is running as root */
	}
}

/*
 * gl_init_config
 *
 * DESCRIPTION: Initialize libconfig.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
gl_init_config(void)
{
	config_init(&cfg);

	if (! config_read_file(&cfg, runtime_files[0])) {
		fprintf(stderr, "%s:%d - %s\n", config_error_file(&cfg),
			config_error_line(&cfg), config_error_text(&cfg));
		config_destroy(&cfg);
		return(EXIT_FAILURE);
	}

	if (LG_VERBOSE == 1) {
		gl_infolog("Initialized libconfig");
	}

	return 0;
}

/*
 * gl_die_config
 *
 * DESCRIPTION: Kill libconfig.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
gl_die_config(void)
{
	config_destroy(&cfg);

	if (LG_VERBOSE == 1) {
		gl_infolog("Destroyed libconfig");
	}

	return(EXIT_SUCCESS);
}

/*
 * gl_load_all_from_config
 *
 * DESCRIPTION: Loads all settings from the Glacier config file.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
gl_load_all_from_config(void)
{
	/* this is probably really ugly but it works */
	if (! config_lookup_bool(&cfg, "GLACIER_DO_INT_CHECK", &GLACIER_DO_INT_CHECK)) { return 1; }
	if (! config_lookup_bool(&cfg, "GLACIER_VERBOSE",      &GLACIER_VERBOSE)) { return 1; }
	
	return 0;
}

/*
 * gl_load_all_from_profile
 *
 * DESCRIPTION: Loads all settings from the Glacier system profile.
 * PARAMETERS: None.
 * DEFINED IN: config.h
 *
 */

int
gl_load_all_from_profile(void)
{
	if (! config_lookup_string(&cfg, "GLACIER_REPO", 	     &GLACIER_REPO)) { return 1; }
	if (! config_lookup_string(&cfg, "GLACIER_ARCH",             &GLACIER_ARCH)) { return 1; }
	if (! config_lookup_string(&cfg, "GLACIER_TARGET", 	     &GLACIER_TARGET)) { return 1; }
	if (! config_lookup_string(&cfg, "GLACIER_SYSTEM_PROFILE",   &GLACIER_SYSTEM_PROFILE)) { return 1; }

	return 0;
}

/*
 * gl_load_setting_from_config
 *
 * DESCRIPTION: Load a specified setting from the Glacier config file.
 * PARAMETERS: char SETTING[]
 * DEFINED IN: config.h
 *
 */

/* int
load_setting_from_config(char SETTING[])
{} */

/*
* gl_create_node
*
* DESCRIPTION: Create a dependency tree node.
* PARAMETERS: char *data
* DEFINED IN: data.h
*
*/

struct node
*gl_create_node(char *data)
{
	if (data == NULL) {
		if (LG_VERBOSE == 1) { gl_errlog("NULL data passed to gl_create_node()"); }
		return NULL;
	}

	struct node *newNode = (struct node*)malloc(sizeof(struct node));
	if (newNode == NULL) {
		if (LG_VERBOSE == 1) { gl_errlog("Memory allocation failed in gl_create_node()"); }
		return NULL;
	}

	newNode->data = strdup(data);
	if (newNode->data == NULL) {
		if (LG_VERBOSE == 1) { gl_errlog("String duplication failed in gl_create_node()"); }
		free(newNode);
		return NULL;
	}

	newNode->numChildren = 0;
	newNode->left = NULL;
	newNode->right = NULL;
	
	for (int i = 0; i < MAX_CHILDREN; i++) {
		newNode->children[i] = NULL;
	}
	
	return newNode;
}

/*
 * gl_free_node
 *
 * DESCRIPTION: Recursively free a node and all its children.
 * PARAMETERS: struct node *root
 * DEFINED IN: data.h
 *
 */

void
gl_free_node(struct node *root)
{
	if (root == NULL) {
		return;
	}
	
	/* Free all children recursively */
	for (int i = 0; i < root->numChildren; i++) {
		gl_free_node(root->children[i]);
	}
	
	/* Free data string */
	if (root->data != NULL) {
		free(root->data);
	}
	
	/* Free node itself */
	free(root);
}

/*
 * gl_add_child
 *
 * DESCRIPTION: Add a child node to a parent node.
 * PARAMETERS: struct node *parent, struct node *child
 * DEFINED IN: data.h
 *
 */

int
gl_add_child(struct node *parent, struct node *child)
{
	if (parent == NULL || child == NULL) {
		if (LG_VERBOSE == 1) { 
			gl_errlog("NULL pointer passed to gl_add_child()"); 
		}
		return 1;
	}

	if (parent->numChildren < 0 || parent->numChildren >= MAX_CHILDREN) {
		if (LG_VERBOSE == 1) { 
			gl_errlog("Invalid numChildren value in parent node");
		}
		return 2;
	}

	if (parent->numChildren < MAX_CHILDREN) {
		parent->children[parent->numChildren++] = child;
		return 0;
	} else {
		if (LG_VERBOSE == 1) { 
			gl_errlog("Maximum number of children exceeded");
		}
		return 3;
	}
}

/*
* gl_print_tree
*
* DESCRIPTION: Print a dependency tree.
* PARAMETERS: struct node *root, int level
* DEFINED IN: data.h
*
*/

int
gl_print_tree(struct node *root, int level)
{
	if (root == NULL) {
		return 0;
	}
	
	if (level < 0) {
		if (LG_VERBOSE == 1) {
			gl_errlog("Invalid level value in gl_print_tree()");
		}
		return 1;
	}
	
	if (level > MAX_RECURSION_DEPTH) {
		/* Safety check to prevent stack overflow from recursive calls */
		if (LG_VERBOSE == 1) {
			gl_errlog("Maximum recursion depth exceeded in gl_print_tree()");
		}
		return 2;
	}

	for (int i = 0; i < level; i++) {
		printf("	");
	}

	if (root->data == NULL) {
		printf("(NULL)\n");
	} else {
		printf("%s\n", root->data);
	}

	if (root->numChildren < 0 || root->numChildren > MAX_CHILDREN) {
		if (LG_VERBOSE == 1) {
			gl_errlog("Invalid numChildren value in node");
		}
		return 3;
	}

	for (int i = 0; i < root->numChildren; i++) {
		if (gl_print_tree(root->children[i], level + 1) != 0) {
			/* Propagate errors up the call stack */
			return 4;
		}
	}
	
	return 0;
}

/*
 * gl_mkworkspace
 *
 * DESCRIPTION: Creates a new Glacier workspace in /tmp.
 * PARAMETERS: None.
 * DEFINED IN: pkgops.h
 *
 */

int
gl_mkworkspace(void)
{
	const char *workspace_path = "/tmp/glacier-workspace";
	
	DIR* workspace = opendir(workspace_path);
	if (workspace) {
		/* Workspace exists */
		closedir(workspace);
		/* gl_infolog("Not creating new workspace, valid workspace already exists."); */
		return 0;
	} else if (ENOENT == errno) {
		/* gl_infolog("Creating new Glacier workspace..."); */
		if (mkdir(workspace_path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH) != 0) {
			if (LG_VERBOSE == 1) {
				gl_errlog("in gl_mkworkspace()");
				gl_errlog("mkdir() failed to create workspace directory");
				fprintf(stderr, "Error: %s\n", strerror(errno));
			}
			return -1;
		}
		return 1;
	} else {
		/* Some other error occurred */
		if (LG_VERBOSE == 1) {
			gl_errlog("in gl_mkworkspace()");
			gl_errlog("opendir() failed to check workspace");
			fprintf(stderr, "Error: %s\n", strerror(errno));
		}
		return -1;
	}
}

/*
 * gl_prepare_pkg
 *
 * DESCRIPTION: Copies a package archive from the localdb to the workspace, and unpacks it.
 * PARAMETERS: char PACKAGE[]
 * DEFINED IN: pkgops.h
 *
 */

int
gl_prepare_pkg(char PACKAGE[])
{
	if (PACKAGE == NULL) {
		printf(COL_RED "[x] " COL_RESET "Package '%s' does not exist in the local database.\n", PACKAGE);
		gl_errlog("Ensure your local database is up to date and try again.");
		gl_errlog("This can be done by running 'glacier-update-pkgdb' as root.");
		return 1;
	}

	/* Get just the package name without path and extension */
	char pkg_name[PATH_MAX_SIZE];
	const char *last_slash = strrchr(PACKAGE, '/');
	const char *last_dot = strrchr(PACKAGE, '.');
	
	if (last_slash == NULL) {
		/* No path separator found, use whole string */
		strncpy(pkg_name, PACKAGE, sizeof(pkg_name) - 1);
	} else {
		/* Copy everything after the last slash */
		strncpy(pkg_name, last_slash + 1, sizeof(pkg_name) - 1);
	}
	
	/* Remove .tar extension if present */
	if (last_dot != NULL && strcmp(last_dot, ".tar") == 0) {
		pkg_name[strlen(pkg_name) - 4] = '\0';
	}

	/* Construct the workspace path */
	char PKG_NEW[PATH_MAX_SIZE];
	snprintf(PKG_NEW, sizeof(PKG_NEW), "/tmp/glacier-workspace/%s.tar", pkg_name);
	
	FILE *pkg_old, *pkg_new;
	char buffer[BUFFER_SIZE];
	size_t bytes_read;

	pkg_old = fopen(PACKAGE, "rb");
	if (pkg_old == NULL) {
		gl_errlog("Failed to open source package file");
		return 1;
	}
	
	pkg_new = fopen(PKG_NEW, "wb");
	if (pkg_new == NULL) {
		gl_errlog("Failed to create destination package file");
		fclose(pkg_old);
		return 1;
	}

	/* Use buffered I/O for better efficiency */
	while ((bytes_read = fread(buffer, 1, sizeof(buffer), pkg_old)) > 0) {
		if (fwrite(buffer, 1, bytes_read, pkg_new) != bytes_read) {
			gl_errlog("Failed to write to destination file");
			fclose(pkg_old);
			fclose(pkg_new);
			return 1;
		}
	}

	fclose(pkg_old);
	fclose(pkg_new);
	
	/* Create the package directory first */
	char pkg_dir[PATH_MAX_SIZE];
	snprintf(pkg_dir, sizeof(pkg_dir), "/tmp/glacier-workspace/%s", pkg_name);
	if (mkdir(pkg_dir, DEFAULT_PERMISSIONS) != 0 && errno != EEXIST) {
		gl_errlog("Failed to create package directory");
		return 1;
	}
	
	/* Change to the package directory before extracting */
	if (chdir(pkg_dir) != 0) {
		gl_errlog("Failed to change to package directory");
		return 1;
	}
	
	/* Fork before exec to prevent process replacement */
	pid_t pid = fork();
	if (pid < 0) {
		gl_errlog("Failed to fork process");
		return 1;
	} else if (pid == 0) {
		/* Child process */
		char *tar_args[] = {
			"/usr/bin/bsdtar", /* Using bsdtar instead of GNU tar */
			"-xf",            /* Extract without verbose output */
			PKG_NEW,
			NULL,
		};

		execvp("/usr/bin/bsdtar", tar_args);
		/* If we get here, execvp failed */
		fprintf(stderr, COL_RED "[x] " COL_RESET "Failed to execute bsdtar: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	} else {
		/* Parent process */
		int status;
		waitpid(pid, &status, 0);
		
		if (WIFEXITED(status) && WEXITSTATUS(status) != 0) {
			printf(COL_RED "[x] " COL_RESET "Error while unpacking archive for package %s.\n", pkg_name);
			return WEXITSTATUS(status);
		}
		
		if (remove(PKG_NEW) != 0) {
			gl_warnlog("Failed to remove temporary package file");
		}
		
		return 0;
	}
	
	/* Should never reach here, but added for completeness */
	return 1;
}

/*
 * gl_run_make_task
 *
 * DESCRIPTION: Runs a make task in current working directory 
 * PARAMETERS: char TASK[]
 * DEFINED IN: pkgops.h
 *
 */

int
gl_run_make_task(char TASK[])
{
	if (TASK == NULL) {
		gl_errlog("No task specified for make");
		return 1;
	}

	/* Fork before exec to prevent process replacement */
	pid_t pid = fork();
	if (pid < 0) {
		gl_errlog("Failed to fork process");
		return 1;
	} else if (pid == 0) {
		/* Child process */
		char *build_args[] = {
			"/bin/make", /* This should be changed to /glacier/bin/make later on */
			TASK,
			NULL
		};

		execvp("/bin/make", build_args);
		/* If we get here, execvp failed */
		gl_errlog("Failed to execute make");
		fprintf(stderr, "Error: %s\n", strerror(errno));
		exit(EXIT_FAILURE);
	} else {
		/* Parent process */
		int status;
		waitpid(pid, &status, 0);
		
		if (WIFEXITED(status)) {
			return WEXITSTATUS(status);
		} else {
			gl_errlog("Make process terminated abnormally");
			return 1;
		}
	}

	/* Should never reach here, but added for completeness */
	return 1;
}

/*
 * gl_print_hash
 *
 * DESCRIPTION: Prints a specified hash string
 * PARAMETERS: unsigned char *hash, unsigned int length
 * DEFINED IN: security.h
 *
 */

int
gl_print_hash(uchar *hash, uint length)
{
	if (hash == NULL) {
		gl_errlog("NULL hash pointer passed to gl_print_hash()");
		return 1;
	}

	if (length == 0) {
		gl_warnlog("Zero-length hash passed to gl_print_hash()");
	}

	for (uint index = 0; index < length; index++) {
		printf("%02x", hash[index]);
	}

	printf("\n");

	return 0;
}

/*
 * gl_stash_hash
 *
 * DESCRIPTION: Stores a hash inside a string
 * PARAMETERS: unsigned char *stored_hash, unsigned char *hash, unsigned int length
 * DEFINED IN: security.h
 *
 */

int
gl_stash_hash(char *stored_hash, unsigned int stored_hash_size, const uchar *hash, uint length)
{
	if (stored_hash == NULL || hash == NULL) {
		if (LG_VERBOSE == 1) {
			gl_errlog("in gl_stash_hash()");
			gl_errlog("NULL pointer provided");
		}
		return 1;
	}
	
	if (stored_hash_size < (length * 2 + 1)) {
		if (LG_VERBOSE == 1) {
			gl_errlog("in gl_stash_hash()");
			gl_errlog("Output buffer too small for hash");
		}
		return 1;
	}

	/* Clear the buffer first */
	stored_hash[0] = '\0';
	
	/* Format each byte as a 2-character hex value */
	for (uint index = 0; index < length; index++) {
		char hex[3];
		snprintf(hex, sizeof(hex), "%02x", hash[index]);
		strncat(stored_hash, hex, stored_hash_size - strlen(stored_hash) - 1);
	}

	return 0;
}

/*
 * gl_hash_file
 *
 * DESCRIPTION: Performs a hashing operation on a file and stores the result
 * PARAMETERS: const char *filename
 * DEFINED IN: security.h
 *
 */

int
gl_hash_file(const char *filename, unsigned char *out_hash, unsigned int *out_length)
{
	FILE *data = fopen(filename, "rb");
	if (! data && LG_VERBOSE == 1) {
		gl_errlog("in gl_hash_file()");
		gl_errlog("error opening file handle");
		return 1;
	}
	else if (! data) {
		return 1;
	}

	EVP_MD_CTX *context = EVP_MD_CTX_new();
	if (! context && LG_VERBOSE == 1) {
		gl_errlog("in gl_hash_file()");
		gl_errlog("error creating envelope context");
		fclose(data);
		return 2;
	}
	else if (! context) {
		fclose(data);
		return 2;
	}

	if (EVP_DigestInit_ex(context, EVP_sha256(), NULL) != 1) {
		if (LG_VERBOSE == 1) {
			gl_errlog("in gl_hash_file()");
			gl_errlog("error initializing digest");
		}
		EVP_MD_CTX_free(context);
		fclose(data);
		return 3;
	}

	// Read and hash file content
	unsigned char buffer[BUFFER_SIZE];
	size_t bytes_read;
	
	while ((bytes_read = fread(buffer, 1, BUFFER_SIZE, data)) > 0) {
		if (EVP_DigestUpdate(context, buffer, bytes_read) != 1) {
			if (LG_VERBOSE == 1) {
				gl_errlog("in gl_hash_file()");
				gl_errlog("error updating digest");
			}
			EVP_MD_CTX_free(context);
			fclose(data);
			return 4;
		}
	}

	if (ferror(data)) {
		if (LG_VERBOSE == 1) {
			gl_errlog("in gl_hash_file()");
			gl_errlog("error reading file");
		}
		EVP_MD_CTX_free(context);
		fclose(data);
		return 5;
	}

	fclose(data);

	if (EVP_DigestFinal_ex(context, out_hash, out_length) != 1) {
		if (LG_VERBOSE == 1) {
			gl_errlog("in gl_hash_file()");
			gl_errlog("error finalizing digest");
		}
		EVP_MD_CTX_free(context);
		return 6;
	}
	
	EVP_MD_CTX_free(context);

	return 0;
}

/*
 * gl_verify_signature
 *
 * Implementation of verify_signature function declared in security.h
 */

int
gl_verify_signature(char PACKAGE[], char SIGNATURE[])
{
    FILE *package_file = NULL;
    FILE *signature_file = NULL;
    
    /* Check if files exist */
    package_file = fopen(PACKAGE, "rb");
    if (package_file == NULL) {
        gl_errlog("Package file not found");
        return 2;
    }
    fclose(package_file);
    
    signature_file = fopen(SIGNATURE, "rb");
    if (signature_file == NULL) {
        gl_errlog("Signature file not found");
        return 2;
    }
    fclose(signature_file);
    
    /* In a real implementation, we would use OpenSSL or GnuPG to verify 
       the signature against the trusted keyring. This is a placeholder implementation. */
    
    /* For testing purposes, we'll just return success */
    gl_successlog("Signature verification successful");
    return 0;
    
    /* 
       Example implementation using GnuPG (would require gpgme library):
       
       gpgme_ctx_t ctx;
       gpgme_error_t err;
       gpgme_data_t sig, text;
       gpgme_verify_result_t result;
       
       // Initialize GPGME context
       gpgme_check_version(NULL);
       err = gpgme_new(&ctx);
       if (err) {
           gpgme_release(ctx);
           gl_errlog("Failed to create GPGME context");
           return 1;
       }
       
       // Open signature and package files
       err = gpgme_data_new_from_file(&sig, SIGNATURE, 1);
       if (err) {
           gpgme_release(ctx);
           gl_errlog("Failed to open signature file");
           return 2;
       }
       
       err = gpgme_data_new_from_file(&text, PACKAGE, 1);
       if (err) {
           gpgme_data_release(sig);
           gpgme_release(ctx);
           gl_errlog("Failed to open package file");
           return 2;
       }
       
       // Verify signature
       err = gpgme_op_verify(ctx, sig, text, NULL);
       gpgme_data_release(sig);
       gpgme_data_release(text);
       
       if (err) {
           gpgme_release(ctx);
           gl_errlog("Verification failed");
           return 1;
       }
       
       // Check verification result
       result = gpgme_op_verify_result(ctx);
       if (!result || !result->signatures) {
           gpgme_release(ctx);
           gl_errlog("No signatures found");
           return 1;
       }
       
       // Check if signature is valid and from a trusted key
       gpgme_signature_t s = result->signatures;
       if (s->status != GPG_ERR_NO_ERROR) {
           gpgme_release(ctx);
           gl_errlog("Invalid signature");
           return 1;
       }
       
       gpgme_release(ctx);
       gl_successlog("Signature verification successful");
       return 0;
    */
}

/*
 * gl_check_integrity
 *
 * Implementation of check_integrity function declared in security.h
 */

int
gl_check_integrity(char PACKAGE[], char EXPECTED_HASH[])
{
    FILE *package_file = NULL;
    unsigned char calculated_hash[EVP_MAX_MD_SIZE];
    unsigned int hash_length;
    char hash_string[EVP_MAX_MD_SIZE * 2 + 1]; /* Each byte becomes 2 hex chars + null terminator */
    
    /* Check if package file exists */
    package_file = fopen(PACKAGE, "rb");
    if (package_file == NULL) {
        gl_errlog("Package file not found");
        return 2;
    }
    fclose(package_file);
    
    /* Calculate hash of the package file */
    if (gl_hash_file(PACKAGE, calculated_hash, &hash_length) != 0) {
        gl_errlog("Failed to calculate hash");
        return 2;
    }
    
    /* Convert binary hash to hex string */
    if (gl_stash_hash(hash_string, sizeof(hash_string), calculated_hash, hash_length) != 0) {
        gl_errlog("Failed to convert hash to string");
        return 2;
    }
    
    /* Compare calculated hash with expected hash */
    if (strcasecmp(hash_string, EXPECTED_HASH) == 0) {
        gl_successlog("Package integrity verified");
        return 0;
    } else {
        gl_warnlog("Package integrity check failed");
        if (GLACIER_VERBOSE) {
            gl_infolog("Expected hash:");
            gl_infolog(EXPECTED_HASH);
            gl_infolog("Calculated hash:");
            gl_infolog(hash_string);
        }
        return 1;
    }
}

/*
 * gl_get_system_profile
 *
 * Implementation of get_system_profile function declared in runtime.h
 */

char *
gl_get_system_profile(void)
{
    /* Initialize configuration if not already done */
    if (cfg.root == NULL) {
        gl_init_config();
        gl_load_all_from_profile();
    }
    
    /* Return the system profile from global variable */
    if (GLACIER_SYSTEM_PROFILE != NULL) {
        return (char *)GLACIER_SYSTEM_PROFILE;
    } else {
        /* Fallback in case the profile is not set */
        gl_warnlog("System profile not found in configuration. Using default.");
        return "x86_64-musl";
    }
}

/*
 * gl_lock_file
 *
 * DESCRIPTION: Locks a specified file using fcntl.
 * PARAMETERS: const char *filepath - Path to the file to lock
 * RETURNS: 0 on success, -1 on failure
 * DEFINED IN: runtime.h
 */

int
gl_lock_file(const char *filepath)
{
    if (filepath == NULL) {
        gl_errlog("NULL filepath provided to gl_lock_file()");
        return -1;
    }

    /* Check if file exists and is regular file */
    struct stat st;
    if (stat(filepath, &st) == -1) {
        if (errno != ENOENT) {
            if (LG_VERBOSE == 1) {
                gl_errlog("Failed to check file status");
                fprintf(stderr, "Error: %s\n", strerror(errno));
            }
            return -1;
        }
    } else if (!S_ISREG(st.st_mode)) {
        gl_errlog("Not a regular file");
        return -1;
    }

    int file_descriptor = open(filepath, O_RDWR | O_CREAT, DEFAULT_PERMISSIONS);
    if (file_descriptor == -1) {
        if (LG_VERBOSE == 1) {
            if (errno == EEXIST) {
                gl_errlog("File already exists but cannot be opened");
            } else {
                gl_errlog("Failed to open file for locking");
            }
            fprintf(stderr, "Error: %s\n", strerror(errno));
        }
        return -1;
    }

    /* Set up signal handlers to ensure lock is released */
    struct sigaction sa;
    sa.sa_handler = NULL; /* Will be set in child process */
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    if (sigaction(SIGTERM, &sa, NULL) == -1 || sigaction(SIGINT, &sa, NULL) == -1) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Failed to set up signal handlers");
            fprintf(stderr, "Error: %s\n", strerror(errno));
        }
        close(file_descriptor);
        return -1;
    }

    struct flock flock;
    flock.l_type = F_WRLCK;    /* Write lock */
    flock.l_whence = SEEK_SET; /* Start from beginning of file */
    flock.l_start = 0;         /* Offset from l_whence */
    flock.l_len = 0;           /* Lock entire file */
    flock.l_pid = getpid();    /* Process ID */

    /* Try to acquire the lock */
    while (fcntl(file_descriptor, F_SETLK, &flock) == -1) {
        if (errno == EINTR) {
            /* Interrupted by signal, try again */
            continue;
        }
        if (errno == EACCES || errno == EAGAIN) {
            if (LG_VERBOSE == 1) {
                gl_errlog("File is already locked by another process");
            }
        } else if (errno == EBADF) {
            if (LG_VERBOSE == 1) {
                gl_errlog("Invalid file descriptor");
            }
        } else {
            if (LG_VERBOSE == 1) {
                gl_errlog("Failed to acquire file lock");
                fprintf(stderr, "Error: %s\n", strerror(errno));
            }
        }
        close(file_descriptor);
        return -1;
    }

    /* Verify the lock was acquired */
    struct flock verify_flock;
    verify_flock.l_type = F_WRLCK;
    verify_flock.l_whence = SEEK_SET;
    verify_flock.l_start = 0;
    verify_flock.l_len = 0;

    if (fcntl(file_descriptor, F_GETLK, &verify_flock) == -1) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Failed to verify lock acquisition");
            fprintf(stderr, "Error: %s\n", strerror(errno));
        }
        close(file_descriptor);
        return -1;
    }

    if (verify_flock.l_type != F_WRLCK || verify_flock.l_pid != getpid()) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Lock verification failed");
        }
        close(file_descriptor);
        return -1;
    }

    return file_descriptor;
}

/*
 * gl_unlock_file
 *
 * DESCRIPTION: Unlocks a specified file using fcntl.
 * PARAMETERS: int file_descriptor - The file descriptor of the locked file
 * RETURNS: 0 on success, 1 on failure
 * DEFINED IN: pkgops.h
 */

int
gl_unlock_file(int file_descriptor)
{
    if (file_descriptor < 0) {
        gl_errlog("Invalid file descriptor provided to gl_unlock_file()");
        return -1;
    }

    /* Verify file descriptor is valid */
    if (fcntl(file_descriptor, F_GETFD) == -1) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Invalid file descriptor");
        }
        return -1;
    }

    /* Verify current lock status */
    struct flock current_flock;
    current_flock.l_type = F_WRLCK;
    current_flock.l_whence = SEEK_SET;
    current_flock.l_start = 0;
    current_flock.l_len = 0;

    if (fcntl(file_descriptor, F_GETLK, &current_flock) == -1) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Failed to check current lock status");
            fprintf(stderr, "Error: %s\n", strerror(errno));
        }
        close(file_descriptor);
        return -1;
    }

    /* Verify we own the lock */
    if (current_flock.l_type != F_WRLCK || current_flock.l_pid != getpid()) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Cannot unlock: lock is held by another process");
        }
        close(file_descriptor);
        return -1;
    }

    struct flock flock;
    flock.l_type = F_UNLCK;    /* Unlock */
    flock.l_whence = SEEK_SET; /* Start from beginning of file */
    flock.l_start = 0;         /* Offset from l_whence */
    flock.l_len = 0;           /* Unlock entire file */
    flock.l_pid = getpid();    /* Process ID */

    /* Try to release the lock */
    while (fcntl(file_descriptor, F_SETLK, &flock) == -1) {
        if (errno == EINTR) {
            /* Interrupted by signal, try again */
            continue;
        }
        if (errno == EBADF) {
            if (LG_VERBOSE == 1) {
                gl_errlog("Invalid file descriptor");
            }
        } else {
            if (LG_VERBOSE == 1) {
                gl_errlog("Failed to release file lock");
                fprintf(stderr, "Error: %s\n", strerror(errno));
            }
        }
        close(file_descriptor);
        return -1;
    }

    /* Verify the lock was released */
    struct flock verify_flock;
    verify_flock.l_type = F_WRLCK;
    verify_flock.l_whence = SEEK_SET;
    verify_flock.l_start = 0;
    verify_flock.l_len = 0;

    if (fcntl(file_descriptor, F_GETLK, &verify_flock) == -1) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Failed to verify lock release");
            fprintf(stderr, "Error: %s\n", strerror(errno));
        }
        close(file_descriptor);
        return -1;
    }

    if (verify_flock.l_type != F_UNLCK) {
        if (LG_VERBOSE == 1) {
            gl_errlog("Lock release verification failed");
        }
        close(file_descriptor);
        return -1;
    }

    close(file_descriptor);
    return 0;
}
